"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

__version__ = "0.0.17"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "rseng"
PACKAGE_URL = "https://github.com/rseng/rseng"
KEYWORDS = "rse,software,research software,criteria,taxonomy"
DESCRIPTION = "criteria and taxonomy for research software engineering"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (
    ("requests", {"min_version": "2.23.0"}),
    ("pyaml", {"min_version": "20.3.1"}),
    ("Jinja2", {"min_version": None}),
)
TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)


ALL_REQUIRES = INSTALL_REQUIRES



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.main import ResearchSoftware
import logging

bot = logging.getLogger("rseng.client")


def main(args, extra):

    # Create a client to interact with
    client = ResearchSoftware(version=args.ver)

    if args.type == "taxonomy":
        files = client.export_taxonomy_markdown(args.path, force=args.force)
        print("\n".join(files))
    elif args.type == "taxonomy-json":
        filename = client.export_taxonomy_json(args.path, force=args.force)
        print(filename)
    elif args.type == "criteria":
        files = client.export_criteria_markdown(args.path, force=args.force)
        print("\n".join(files))
    elif args.type == "criteria-annotation-template":
        filename = client.export_criteria_annotation_template(
            args.path or "annotate-criteria-template.md", force=args.force
        )
        print(filename)
    elif args.type == "taxonomy-annotation-template":
        filename = client.export_taxonomy_annotation_template(
            args.path or "annotate-taxonomy-template.md", force=args.force
        )
        print(filename)
    else:
        print("Please specify an export type as the first argument.")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.main.taxonomy import Taxonomy
from rseng.main.criteria import CriteriaSet
from rseng.main.templates import get_template
from rseng.utils.file import write_file, write_json

from datetime import datetime
from jinja2 import Template
from copy import deepcopy
import logging
import os
import sys

bot = logging.getLogger("rseng.main")
now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")


class ResearchSoftware:
    """The Research Engineering client will load current taxonomy and criteria
    for research software, and generate documentation files for them.
    """

    def __init__(self, version="latest"):
        """create a software repository. We take a config file, which should
        sit at the root of the repository, and then parse the subfolders
        accordingly.
        """
        self.taxonomy = Taxonomy(version=version)
        self.criteria = CriteriaSet(version=version)

    # Taxonomy Export

    def export_taxonomy_markdown(
        self, outdir=None, template="taxonomy-metadata-template.md", force=False
    ):
        """Given an output directory and a template name, export a set of
        markdown files for the entire taxonomy
        """
        # If output folder not provided, assume _taxonomy in $PWD
        outdir = outdir or "_taxonomy"
        template = get_template(template)

        # If the output file exists and force is false, exit early
        if not force and os.path.exists(outdir):
            sys.exit(f"{outdir} exists, use --force to overwrite.")
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        # Flatten the tree, print each node to file
        files = []
        for path, node in self.taxonomy.flatten().items():
            instance = Template(template)
            result = instance.render(
                name=node.get("name", ""),
                path=path,
                updated_at=now,
                color=node.get("color", ""),
                uid=node.get("uid", ""),
                example=node.get("example"),
                content=path,
            )
            output = os.path.join(outdir, "%s.md" % node.get("uid"))
            write_file(output, result)
            files.append(output)

        return files

    def export_taxonomy_json(self, outfile=None, force=False, size=12):
        """Given an output json file, export the taxonomy d3 data. In the future,
        sizes could be customized based on importance, etc.
        """
        outfile = outfile or "taxonomy.json"

        # If the output file exists and force is false, exit early
        if not force and os.path.exists(outfile):
            sys.exit(f"{outfile} exists, use --force to overwrite.")

        # Make a copy of nodes before editing
        tree = deepcopy(self.taxonomy.tree)
        for uid, node in self.taxonomy.nodes.items():
            node["size"] = size

        # Generate the tree in the typical d3.js format
        root = {"name": "root", "children": self.taxonomy.tree}

        # Restore original tree (without sizes)
        self.taxonomy.tree = tree
        return write_json(root, outfile)

    def export_taxonomy_annotation_template(
        self, outfile, template="annotate-taxonomy-template.md", force=False
    ):
        """Export a markdown template for a taxonomy annotation (as GitHub issue)"""
        items = ["%s\n%s\n" % (v["uid"], k) for k, v in self.taxonomy.flatten().items()]
        return self.export_annotation_template(
            template=template, items=items, outfile=outfile, force=force
        )

    # Annotation template

    def export_annotation_template(self, template, items, outfile=None, force=False):
        """Given an output directory and a template name, export a set of
        markdown files for each criteria or taxonomy item.
        """
        # If output folder not provided, assume _taxonomy in $PWD
        template = get_template(template)

        # If the output file exists and force is false, exit early
        if outfile and not force and os.path.exists(outfile):
            sys.exit(f"{outfile} exists, use --force to overwrite.")

        instance = Template(template)
        result = instance.render(items=items)
        if outfile:
            write_file(outfile, result)
            return outfile
        return result

    # Criteria Export

    def export_criteria_markdown(
        self, outdir, template="criteria-metadata-template.md", force=False
    ):
        """Given an output directory and a template name, export a set of
        markdown files for each criteria
        """
        # If output folder not provided, assume _taxonomy in $PWD
        outdir = outdir or "_criteria"
        template = get_template(template)

        # If the output file exists and force is false, exit early
        if not force and os.path.exists(outdir):
            sys.exit(f"{outdir} exists, use --force to overwrite.")
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        files = []
        for criteria in self.criteria:
            instance = Template(template)
            result = instance.render(
                name=criteria.question,
                updated_at=now,
                uid=criteria.uid,
                options=criteria.options,
            )
            output = os.path.join(outdir, "%s.md" % criteria.uid)
            write_file(output, result)
            files.append(output)

        return files

    def export_criteria_annotation_template(
        self, outfile, template="annotate-criteria-template.md", force=False
    ):
        """Given an output directory and a template name, export a set of
        markdown files for each criteria
        """
        items = ["%s\n%s\n" % (c.uid, c.question) for c in self.criteria]
        return self.export_annotation_template(
            template=template, items=items, outfile=outfile, force=force
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.utils.file import read_file
import os

here = os.path.dirname(os.path.abspath(__file__))


def get_template(name, load=True):
    """Given the name of a template (a markdown file to use to generate jekyll
    markdown frontend matter) return the template path if it exists.
    """
    # The user can provide a custom template
    if not os.path.exists(name):
        filename = os.path.join(here, name)
    else:
        filename = name

    if not os.path.exists(filename):
        raise FileNotFoundError(f"{filename} does not exist.")

    # Does the user want to load it?
    if load:
        filename = read_file(filename, readlines=False)
    return filename



#!/usr/bin/env python

"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.logger import RSENG_LOG_LEVELS, RSENG_LOG_LEVEL
import rseng
import argparse
import sys
import logging


def get_parser():
    parser = argparse.ArgumentParser(
        description="Research software engineering taxonomy and criteria."
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--log_level",
        dest="log_level",
        choices=RSENG_LOG_LEVELS,
        default=RSENG_LOG_LEVEL,
        help="Customize logging level for rse inspector.",
    )

    description = "actions for rseng"
    subparsers = parser.add_subparsers(
        help="rseng actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", help="show software version")

    # Generate a key for the interface
    generate = subparsers.add_parser(
        "generate",
        help="generate output files for taxonomy and criteria",
    )

    generate.add_argument(
        "type",
        help="Type to generate",
        choices=[
            "taxonomy",
            "criteria",
            "taxonomy-json",
            "criteria-annotation-template",
            "taxonomy-annotation-template",
        ],
        nargs="?",
    )

    generate.add_argument(
        "path",
        help="Path (directory) to generate files into (should not exist)",
        nargs="?",
        default=None,
    )

    generate.add_argument(
        "ver",
        help="Version of criteria and taxonomy to use",
        nargs="?",
        default="latest",
    )

    generate.add_argument(
        "--force",
        dest="force",
        help="force generation of folder if already exists.",
        default=False,
        action="store_true",
    )

    return parser


def main():
    """main entrypoint for rse"""

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """
        version = rseng.__version__

        print("\nResearch Software Engineering taxonomy and criteria v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Set the logging level
    logging.basicConfig(level=getattr(logging, args.log_level))
    bot = logging.getLogger("rseng.client")
    bot.setLevel(getattr(logging, args.log_level))

    # Show the version and exit
    if args.command == "version" or args.version:
        print(rseng.__version__)
        sys.exit(0)

    # Does the user want a shell?
    main = None
    if args.command == "generate":
        from .generate import main

    if main is not None:
        main(args=args, extra=extra)
    else:
        help()


if __name__ == "__main__":
    main()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.utils.file import read_yaml, write_file
import logging
import os
import sys

here = os.path.dirname(os.path.abspath(__file__))
bot = logging.getLogger("rseng.main.taxonomy")


class Taxonomy:
    """a taxonomy holds the research software taxonomy"""

    def __init__(self, data_file=None, version="latest"):
        """data_file, if provided, should be a custom owx (owl xml file)
        that is typically generated via export from webprotege.stanford.edu

        Arguments:
          data_file (str) : path to a data file, with ids, questions, options
          version (str)   : version string for file criteria-<version>.tsv
        """
        self.tree = None
        self.nodes = {}

        if not data_file:
            data_file = os.path.join(here, "taxonomy-%s.yaml" % version)

        if not os.path.exists(data_file):
            raise FileNotFoundError

        self.load(data_file)

    @property
    def count(self):
        return len(self.nodes)

    def export(self, filename=None, force=True, sep="\t"):
        """save taxonomy to human readable file"""
        if filename is not None and os.path.exists(filename) and force is False:
            sys.exit("{filename} exists! Set force=True to override it.")

        lines = []
        for uid, criteria in self.criteria.items():
            lines.append(criteria.export(sep))
        if not filename:
            return "\n".join(lines)
        write_file(filename, "\n".join(lines))
        return filename

    def __str__(self):
        return "[Taxonomy:%s]" % len(self.nodes)

    def __repr__(self):
        return self.__str__()

    def flatten(self):
        out = {}

        def flatten(x, name=""):
            if isinstance(x, dict):
                if "children" in x:
                    if not name:
                        flatten(x["children"], x["name"])
                    else:
                        flatten(x["children"], name + " >> " + x["name"])
                else:
                    out[name + " >> " + x["name"]] = x

            elif isinstance(x, list):
                for i, elem in enumerate(x):
                    flatten(elem, name)

        flatten(self.tree)
        return out

    def load(self, data_file):
        """Load a data file"""
        self.tree = read_yaml(data_file)
        nodes = self.tree.copy()
        while nodes:
            node = nodes.pop(0)

            # Add node to lookup, validate as we go
            if node["uid"] not in self.nodes:
                self.nodes[node["uid"]] = node
                assert "name" in node and "uid" in node
                assert node["uid"].startswith("RSE-taxonomy")
            nodes += node.get("children", [])



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rseng.utils.file import read_file, write_file

import logging
import os
import sys

here = os.path.dirname(os.path.abspath(__file__))
bot = logging.getLogger("rseng.main.criteria")


class CriteriaSet:
    """a criteria set holds and managed a group of criteria"""

    def __init__(self, data_file=None, sep="\t", version="latest"):
        """data_file should be a delimited file that has unique ids, questions,
        and comma separated answer options. By default, we use the latest
        version.

        Arguments:
          data_file (str) : path to a data file, with ids, questions, options
          sep (str)       : delimiter to use for the data file
          version (str)   : version string for file criteria-<version>.tsv
        """
        self.criteria = dict()
        if not data_file:
            data_file = os.path.join(here, "criteria-%s.tsv" % version)

        if not os.path.exists(data_file):
            raise FileNotFoundError

        self.load(data_file, sep)

    @property
    def count(self):
        return len(self.criteria)

    def export(self, filename=None, force=True, sep="\t"):
        """save criteria to a file."""
        if filename is not None and os.path.exists(filename) and force is False:
            sys.exit("{filename} exists! Set force=True to override it.")

        lines = []
        for uid, criteria in self.criteria.items():
            lines.append(criteria.export(sep))
        if not filename:
            return "\n".join(lines)
        write_file(filename, "\n".join(lines))
        return filename

    def __str__(self):
        return "[CriteriaSet:%s]" % len(self.criteria)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        """Allow the user to iterate over criteria."""
        for name, criteria in self.criteria.items():
            yield criteria

    def load(self, data_file, sep="\t"):
        """Load a data file with a particular separator."""
        for line in read_file(data_file):
            try:
                uid, question, options = line.strip().split(sep)
                self.criteria[uid] = Criteria(uid, question, options.split(","))
            except:
                bot.warning(f"Malformed line {line}, skipping.")


class Criteria:
    """A parser base exists to extract and format repository metadata."""

    def __init__(self, uid, question, options="yes,no"):
        """a criteria includes a unique id, question, and options."""
        self.uid = uid
        self.question = question
        if isinstance(options, str):
            options = options.split(",")
        self.options = options

    def export(self, sep="\t"):
        """Given a separator, export a criteria to a line to write to file"""
        return "{}{}{}{}{}".format(
            self.uid, sep, self.question, sep, ",".join(self.options)
        )

    def __str__(self):
        return "[Criteria:%s,%s]" % (self.uid, self.question)

    def __repr__(self):
        return self.__str__()

    def summary(self):
        return "[%s][%s]" % (self.uid, self.question)



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_criteria(tmp_path):
    """test criteria sets
    """
    print("Testing Criteria Sets")
    from rseng.main.criteria import CriteriaSet

    cset = CriteriaSet()
    assert cset.count >= 6
    assert len(cset.criteria) >= 6
    assert cset.count == len(cset.criteria)
    for name, criteria in cset.criteria.items():
        assert name.startswith("RSE-")
        assert criteria.question
        assert criteria.uid == name
        assert criteria.export()
    assert cset.export()
    list(cset)
    for criteria in cset:
        print(criteria)


def test_taxonomy(tmp_path):
    """test taxonomy items
    """
    print("Testing Taxonomy items")
    from rseng.main.taxonomy import Taxonomy

    tax = Taxonomy()
    assert tax.count >= 24
    for name in tax.flatten():
        print(name)
    seen_colors = set()
    for name, item in tax.flatten().items():
        for key in ["uid", "name", "color"]:
            assert key in item
        assert item["color"] not in seen_colors
        seen_colors.add(item["color"])
